/* $Id: kgimon.c,v 1.6 1999/02/05 00:59:51 marcus Exp $
******************************************************************************

   Misc routines to deal with KGI monitor databases
   
   Copyright (C) 1998-1999 Marcus Sundberg	[marcus@ggi-project.org]
   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <kgi/kgi.h>
#include <kgi/kgi_commands.h>

#include "kgimon.h"

int
skip_white(FILE *fp)
{
	int c;
	
	do {
		c = fgetc(fp);
		if (c == EOF) {
			return EOF;
		}
	} while (isspace(c));

	ungetc(c, fp);
	return c;
}

int
skip_line(FILE *fp)
{
	int c;
	
	do {
		c = fgetc(fp);
	} while (c != EOF && c != '\n');

	return c;
}

int
skip_comments(FILE *fp)
{
	int c;
	
	do {
		if (skip_white(fp) == EOF) {
			return EOF;
		}
		c = fgetc(fp);
		if (c != COMMENT_CHAR)
			break;
		if (skip_line(fp) == EOF) {
			return EOF;
		}
	} while (1);
	
	if (c != EOF) {
		ungetc(c, fp);
	}
	return c;
}


char *
str_stripwhite(char *str)
{
	char *strend;

	while (isspace(*str)) {
		str++;
	}
	strend = str;
	while ((!isspace(*strend)) && *strend != '\0') {
		strend++;
	}
	*strend = '\0';
	
	return str;
}

kgi_graphmask
parse_graphmask(char *str)
{
	/* Are there any monitors that cares about the depth???
	   The current graphtype system is obsolete anyway so
	   we return GM_ALL for everything */
	return GM_ALL;
}

enum kgi_flags
parse_kgiflags(char *str)
{
	enum kgi_flags flags = 0;
	char *s, *endptr;

	if (strlen(str) < 1) return FL_MON_NORMAL;
	
	s = strtok(str, "|");
	do {
		s = str_stripwhite(s);
		if (strcmp(s, "FL_MON_NORMAL") == 0) {
			flags |= FL_MON_NORMAL;
		} else if (strcmp(s, "FL_MON_RIGHT_EYE") == 0) {
			flags |= FL_MON_RIGHT_EYE;
		} else if (strcmp(s, "FL_MON_LEFT_EYE") == 0) {
			flags |= FL_MON_LEFT_EYE;
		} else if (strcmp(s, "FL_MON_STEREO") == 0) {
			flags |= FL_MON_STEREO;
		} else if (strcmp(s, "FL_MON_INTERLACE") == 0) {
			flags |= FL_MON_INTERLACE;
		} else if (strcmp(s, "FL_MON_POWERSAVE") == 0) {
			flags |= FL_MON_POWERSAVE;
		} else if (strcmp(s, "FL_MON_PROPSIZE") == 0) {
			flags |= FL_MON_PROPSIZE;
		}
	} while ((s = strtok(NULL, "|")) != NULL);

	if (flags == 0) {
		flags = strtol(str, &endptr, 0);
		if (endptr == str) {
			return 0;
		}
	}
	return flags;
}

enum kgi_display_type
parse_displaytype(char *str)
{
	str = str_stripwhite(str);
	
	if (strcmp(str, "MONO_DISPLAY") == 0) {
		return MONO_DISPLAY;
	}
	if (strcmp(str, "HERCMONO") == 0) {
		return HERCMONO;
	}
	if (strcmp(str, "LCD") == 0) {
		return LCD;
	}
	if (strcmp(str, "VGA_MONO") == 0) {
		return VGA_MONO;
	}
	if (strcmp(str, "RGB_DISPLAY") == 0) {
		return RGB_DISPLAY;
	}
	if (strcmp(str, "EGA_8") == 0) {
		return EGA_8;
	}
	if (strcmp(str, "EGA_16") == 0) {
		return EGA_16;
	}
	if (strcmp(str, "CGA") == 0) {
		return CGA;
	}
	if (strcmp(str, "VGA_COLOR") == 0) {
		return VGA_COLOR;
	}
	if (strcmp(str, "PLASMA") == 0) {
		return PLASMA;
	}
	if (strcmp(str, "TFT") == 0) {
		return TFT;
	}
	if (strcmp(str, "COLOR_LCD") == 0) {
		return COLOR_LCD;
	}
	if (strcmp(str, "NON_RGB_DISPLAY") == 0) {
		return NON_RGB_DISPLAY;
	}

	return strtol(str, NULL, 0);
}

enum kgi_sync
parse_synctype(char *str)
{
	enum kgi_sync sync = 0;
	char *s, *endptr;
	
	if (strlen(str) < 1) return SYNC_NORMAL;
	
	s = strtok(str, "|");
	do {
		s = str_stripwhite(s);
		if (strcmp(s, "SYNC_SEPARATE") == 0) {
			sync |= SYNC_SEPARATE;
		} else if (strcmp(s, "SYNC_ON_RED") == 0) {
			sync |= SYNC_ON_RED;
		} else if (strcmp(s, "SYNC_ON_GREEN") == 0) {
			sync |= SYNC_ON_GREEN;
		} else if (strcmp(s, "SYNC_ON_BLUE") == 0) {
			sync |= SYNC_ON_BLUE;
		} else if (strcmp(s, "SYNC_SERRATE") == 0) {
			sync |= SYNC_SERRATE;
		} else if (strcmp(s, "SYNC_COMPOSITE") == 0) {
			sync |= SYNC_COMPOSITE;
		} else if (strcmp(s, "SYNC_NORMAL") == 0) {
			sync |= SYNC_NORMAL;
		} else if (strcmp(s, "SYNC_PEDESTRAL") == 0) {
			sync |= SYNC_PEDESTRAL;
		} else if (strcmp(s, "SYNC_0700_0300") == 0) {
			sync |= SYNC_0700_0300;
		} else if (strcmp(s, "SYNC_0714_0286") == 0) {
			sync |= SYNC_0714_0286;
		} else if (strcmp(s, "SYNC_0100_0400") == 0) {
			sync |= SYNC_0100_0400;
		} else if (strcmp(s, "SYNC_UNKNOWNLEVEL") == 0) {
			sync |= SYNC_UNKNOWNLEVEL;
		} else if (strcmp(s, "SYNC_STANDBY") == 0) {
			sync |= SYNC_STANDBY;
		} else if (strcmp(s, "SYNC_SUSPEND") == 0) {
			sync |= SYNC_SUSPEND;
		} else if (strcmp(s, "SYNC_OFF") == 0) {
			sync |= SYNC_OFF;
		} else if (strcmp(s, "SYNC_VESA_DPMS") == 0) {
			sync |= SYNC_VESA_DPMS;
		} else if (strcmp(s, "SYNC_SSYNC_RIGHT") == 0) {
			sync |= SYNC_SSYNC_RIGHT;
		} else if (strcmp(s, "SYNC_SSYNC_LEFT") == 0) {
			sync |= SYNC_SSYNC_LEFT;
		} else if (strcmp(s, "SYNC_MULTISYNC") == 0) {
			sync |= SYNC_MULTISYNC;
		}
	} while ((s = strtok(NULL, "|")) != NULL);

	if (sync == 0) {
		sync = strtol(str, &endptr, 0);
		if (endptr == str) {
			return 0;
		}
	}
	return sync;
}

kgiu16
parse_polarity(char *str)
{
	kgiu16 pol = 0;
	char *s, *endptr;
	
	if (strlen(str) < 1) return 0;
	
	s = strtok(str, "|");
	do {
		s = str_stripwhite(s);
		if (strcmp(s, "POL_HPOS") == 0) {
			pol |= POL_HPOS;
		} else if (strcmp(s, "POL_HNEG") == 0) {
			pol |= POL_HNEG;
		} else if (strcmp(s, "POL_VPOS") == 0) {
			pol |= POL_VPOS;
		} else if (strcmp(s, "POL_VNEG") == 0) {
			pol |= POL_VNEG;
		}
	} while ((s = strtok(NULL, "|")) != NULL);

	if (pol == 0) {
		pol = strtol(str, &endptr, 0);
		if (endptr == str) {
			return 0;
		}
	}
	return pol;
}	

void
fprint_coord(FILE *fp, char *name, kgi_coord *coord)
{
	printf("%s%d x %d\n", name, coord->x, coord->y);
}

int
fprint_monitor(FILE *fp, struct kgi_monitor *mon)
{
	fprintf     (fp, "\nKGI Monitor Info\n");
	fprintf     (fp, "  Manufacturer  : %s\n", mon->manufact);
	fprintf     (fp, "  Model         : %s\n", mon->model);
	fprintf     (fp, "  Max resolution: %d x %d\n",
		    mon->maxdots.x, mon->maxdots.y);
	fprint_coord(fp, "  Size (in mm)  : ", &mon->size);
	fprintf     (fp, "  Bandwidth     : %3d - %3d MHz\n",
		    mon->dclk.min/1000000, mon->dclk.max/1000000);
	fprintf     (fp, "  HFreq         : %3d - %3d KHz\n",
		     mon->hfreq.min/1000, mon->hfreq.max/1000);
	fprintf     (fp, "  VFreq         : %3d - %3d Hz\n",
		    mon->vfreq.min, mon->vfreq.max);
	fprintf     (fp, "\n");

	return 0;
}

int
read_dbentry(FILE *fp, struct kgi_monitor *monp, struct kgi_preset **pres)
{
	int i;
	char line[MAX_LINE];
	
	if (skip_white(fp) == EOF) {
		return END_OF_FILE;
	}
	/* Vendor name */
	if (fgets(line, MAX_LINE, fp) == NULL) {
		return END_OF_FILE;
	}
	for (i = 0; i < MAX_NAMELEN && line[i] != '\0' && line[i] != '\n';
	     i++) {
		monp->manufact[i] = line[i];
	}
	monp->manufact[i] = '\0';

	/* Model name */
	if (fgets(line, MAX_LINE, fp) == NULL) {
		return END_OF_FILE;
	}
	for (i = 0; i < MAX_NAMELEN && line[i] != '\0' && line[i] != '\n';
	     i++) {
		monp->model[i] = line[i];
	}
	monp->model[i] = '\0';

	/* Supported modes */
	if (fgets(line, MAX_LINE, fp) == NULL) {
		return END_OF_FILE;
	}
	monp->modes = parse_graphmask(line);
	
	/* Flags */
	if (fgets(line, MAX_LINE, fp) == NULL) {
		return END_OF_FILE;
	}
	monp->flags = parse_kgiflags(line);
	
	/* Max resolution and physical size*/
	if (fscanf(fp, " %hd x %hd \n %hd x %hd \n ",
		   &monp->maxdots.x, &monp->maxdots.y,
		   &monp->size.x, &monp->size.y) != 4) {
		return PARSE_ERROR;
	}
	
	/* Display type */
	if (fgets(line, MAX_LINE, fp) == NULL) {
			return END_OF_FILE;
	}
	monp->disptype = parse_displaytype(line);

	/* Sync type */
	if (fgets(line, MAX_LINE, fp) == NULL) {
			return END_OF_FILE;
	}
	monp->synctype = parse_synctype(line);
	
	/* Bandwidth, H/V frequency */
	if (fscanf(fp, " %d - %d \n %d - %d \n %d - %d \n ",
		   &monp->dclk.min, &monp->dclk.max,
		   &monp->hfreq.min, &monp->hfreq.max,
		   &monp->vfreq.min, &monp->vfreq.max) != 6) {
		return PARSE_ERROR;
	}
	
	/* Timings */
	*pres = NULL;
	i = 0;
	while (1) {
		if ((*pres = realloc(*pres, sizeof(struct kgi_preset)*(i+2)))
		    == NULL) {
			perror("NO MEMORY");
			exit(10);
		}
		if (fscanf(fp, " %d \n "
			   "%hu , %hu , %hu , %hu , %hu , %hu , %hu \n "
			   "%hu , %hu , %hu , %hu , %hu , %hu , %hu \n ",
			   &((*pres)[i]).dclk,
			   &((*pres)[i]).x.width, &((*pres)[i]).x.blankstart,
			   &((*pres)[i]).x.syncstart, &((*pres)[i]).x.syncend,
			   &((*pres)[i]).x.blankend, &((*pres)[i]).x.total,
			   &((*pres)[i]).x.polarity,
			   &((*pres)[i]).y.width, &((*pres)[i]).y.blankstart,
			   &((*pres)[i]).y.syncstart, &((*pres)[i]).y.syncend,
			   &((*pres)[i]).y.blankend, &((*pres)[i]).y.total,
			   &((*pres)[i]).y.polarity) != 15) {
			break;
		}
		i++;
	}
	((*pres)[i]).dclk = 0;

	return 0;
}


#ifdef HAVE_SET_MONSTUFF
int
get_curtiming(int fd, struct kgi_preset *pres)
{
	int ret;

	if ((ret = ioctl(fd, MONITOR_GETCURTIMING, (void*) pres)) != 0) {
		perror("getting current timing");
	}
	return ret;
}

int
set_monitor(int fd, struct kgi_monitor *mon)
{
	int ret;

	if ((ret = ioctl(fd, MONITOR_SETINFO, (void*) mon)) != 0) {
		perror("setting KGI monitor info");
	}
	return ret;
}

int
set_preset(int fd, struct kgi_preset *pres)
{
	int ret;

	if ((ret = ioctl(fd, MONITOR_SETPRESETMODE, pres)) != 0) {
		perror("setting preset for KGI timelist monitor");
	}
	return ret;
}

int
clr_presets(int fd)
{
	int ret;

	if ((ret = ioctl(fd, MONITOR_CLEARPRESETS)) != 0) {
		perror("clearing presetlist for KGI timelist monitor");
	}
	return ret;
}

int
set_presets(int fd, struct kgi_preset *pres)
{
	int ret, i;
	
	for (i = 0; pres[i].dclk != 0; i++) {
		if ((ret = set_preset(fd, pres+i)) != 0) {
			return ret;
		}
	}
	return 0;
}

int
set_htiming(int fd, struct kgi_timing *htim)
{
	int ret;

	if ((ret = ioctl(fd, MONITOR_SETHTIMING, htim)) != 0) {
		perror("setting htiming for KGI monosync monitor");
	}
	return ret;
}

int
set_vtiming(int fd, struct kgi_timing *vtim)
{
	int ret;

	if ((ret = ioctl(fd, MONITOR_SETVTIMING, vtim)) != 0) {
		perror("setting vtiming for KGI monosync monitor");
	}
	return ret;
}

int
clr_timings(int fd)
{
	int ret;

	if ((ret = ioctl(fd, MONITOR_CLEARTIMINGS)) != 0) {
		perror("clearing timinglist for KGI monosync monitor");
	}
	return ret;
}

int
set_timings(int fd, struct kgi_timing *htim, struct kgi_timing *vtim)
{
	int ret, i;
	
	for (i = 0; htim[i].width != 0; i++) {
		if ((ret = set_htiming(fd, htim+i)) != 0) {
			return ret;
		}
	}
	for (i = 0; vtim[i].width != 0; i++) {
		if ((ret = set_vtiming(fd, vtim+i)) != 0) {
			return ret;
		}
	}
	return 0;
}

int
set_monfromfile(FILE *fp)
{
	struct kgi_monitor mon;
	struct kgi_preset *pres;
	char line[MAX_LINE];
	char line2[MAX_LINE];
	int i, j;
	int fd;
	
	for (i = 0;;i++) {
		memset(&mon, 0, sizeof(mon));
		if (skip_comments(fp) == EOF) {
			return END_OF_FILE;
		}
		if (fgets(line, MAX_LINE, fp) == NULL) {
			printf("NR1\n");
			return PARSE_ERROR;
		}
		for (j = 0; line[j] != ':' && line[j] != '\0'; j++)
			;
		if (line[j] != ':') {
			printf("NR2\n");
			return PARSE_ERROR;
		}
		line[j] = '\0';

		if ((fd = open(line, O_WRONLY)) < 0) {
			sprintf(line2, "unable to open %s", line);
			perror(line2);
		}
		if (read_dbentry(fp, &mon, &pres) != 0) {
			printf("NR3\n");
			return PARSE_ERROR;
		}

		if (fd >= 0) {
#if 0
			fprint_monitor(stdout, &mon);
#endif
			fprintf(stdout, "Setting monitor on %s to %s %s...\n",
			       line, mon.manufact, mon.model);
			
			set_monitor(fd, &mon);
			clr_presets(fd);
			set_presets(fd, pres);
			close(fd);
		}
		if (skip_comments(fp) == EOF) {
			return 0;
		}
	}
	return 0;
}
#endif
